/*  Copyright 2003 Guillaume Duhamel

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "saturn.hh"
#include "vdp1.hh"
#include "intc.hh"
#include "exception.hh"
#include "scu.hh"
#include "timer.hh"

Vdp1Registers::Vdp1Registers(Memory *mem, Scu *scu) : LoggedMemory("vdp1", 0x18) {
  vdp1 = new Vdp1(this, mem, scu);
  vdp1Thread = SDL_CreateThread((int (*)(void*)) &Vdp1::lancer, vdp1);
}

Vdp1Registers::~Vdp1Registers(void) {
#if DEBUG
  cerr << "stopping vdp1\n";
#endif
  vdp1->stop();
  SDL_WaitThread(vdp1Thread, NULL);
#if DEBUG
  cerr << "vdp1 stopped\n";
#endif
  delete vdp1;
}

Vdp1 *Vdp1Registers::getVdp1(void) {
  return vdp1;
}

Vdp1::Vdp1(Vdp1Registers *reg, Memory *mem, Scu *s) {
  memoire = mem;
  registres = reg;
  scu = s;

  registres->setWord(0x4, 0);

#if HAVE_LIBSDL_GFX
  logo = SDL_LoadBMP("logo.bmp");
  SDL_WM_SetIcon(logo, NULL);
  string title("Yabause ");
  title += VERSION;
  SDL_WM_SetCaption(title.c_str(), NULL);
  surface = SDL_SetVideoMode(400,400,16,SDL_DOUBLEBUF|SDL_HWSURFACE);
#endif
}

Vdp1::~Vdp1(void) {
#if HAVE_LIBSDL_GFX
  SDL_FreeSurface(surface);
  SDL_FreeSurface(logo);
#endif
}

void Vdp1::stop(void) {
  _stop = true;
}

void Vdp1::lancer(Vdp1 *vdp1) {
  Timer t;
  while(!vdp1->_stop) {
    t.waitVBlankOUT();
    vdp1->executer();
  }
}

void Vdp1::executer(unsigned long adr) {
  unsigned short commande = memoire->getWord(adr);
  unsigned long suivante;

  if (!registres->getWord(0x4)) return;

  // beginning of a frame (ST-013-R3-061694 page 53)
  // BEF <- CEF
  // CEF <- 0
  registres->setWord(0x10, (registres->getWord(0x10) & 2) >> 1);

  while (!(commande & 0x8000)) {

    //cerr << "vdp1 : commande = " << commande << '\n';
    //cerr << "vdp1 :   | mode de saut = " << ((commande & 0x7000) >> 12) << '\n';

    // First, process the command
    if (!(commande & 0x7000)) { // if (!skip)
      switch (commande & 0x001F) {
	case 0: //cerr << "vdp1\t: normal sprite draw" << endl;
	  normalSpriteDraw();
	  break;
	case 1: //cerr << "vdp1\t:  scaled sprite draw" << endl;
	  scaledSpriteDraw();
	  break;
	case 2: // distorted sprite draw
	  distortedSpriteDraw();
	  break;
	case 4: // polygon draw
	  polygonDraw(adr);
	  break;
	case 5: // polyline draw
	  polylineDraw();
	  break;
	case 6: // line draw
	  lineDraw();
	  break;
	case 8: // user clipping coordinates
	  userClipping();
	  break;
	case 9: // system clipping coordinates
	  systemClipping();
	  break;
	case 10: // local coordinate
	  localCoordinate();
	  break;
	case 0x10: // draw end
	  drawEnd();
	  break;
	default:
	  //cerr << "vdp1\t: Mauvaise commande : "
	  //     << hex << setw(10) << commande << endl;
	  break;
      }
    }

    // Next, determine where to go next
    switch ((commande & 0x3000) >> 12) {
    case 0: // NEXT, jump to following table
      adr += 0x20;
      break;
    case 1: // ASSIGN, jump to CMDLINK
      adr = memoire->getWord(adr + 2) * 8;
      break;
    case 2: // CALL, call a subroutine
      executer(adr + 0x20);
      adr = memoire->getWord(adr + 2) * 8;
      break;
    case 3: // RETURN, return from subroutine
      return;
    }
    //cerr << "vdp1 :   | table suivante = " << adr << '\n';
    try { commande = memoire->getWord(adr); }
    catch (MauvaisAcces ma) { return; }
  }
  // FIXME : on a fini, on place deux bits  1 dans un registre
  registres->setWord(0x10, registres->getWord(0x10) | 2);
#if DEBUG
  cerr << "vdp1 : draw end\n";
#endif
  scu->sendDrawEnd();
}

void Vdp1::normalSpriteDraw(void) {
#if DEBUG
  cerr << "vdp1\t: normal sprite draw" << endl;
#endif
}

void Vdp1::scaledSpriteDraw(void) {
#if DEBUG
  cerr << "vdp1\t:  scaled sprite draw" << endl;
#endif
}

void Vdp1::distortedSpriteDraw(void) {
#if DEBUG
  cerr << "vdp1\t: distorted sprite draw" << endl;
#endif
}

void Vdp1::polygonDraw(unsigned short adr) {
#if DEBUG
  cerr << "vdp1\t: polygon draw" << endl;
#endif
#if HAVE_LIBSDL_GFX
  Sint16 x[4];
  Sint16 y[4];
  unsigned short CMDCOLR = memoire->getWord(adr + 0x6);
  x[0] = memoire->getWord(adr + 0x0C) & 0x7FF;
  y[0] = memoire->getWord(adr + 0x0E) & 0x7FF;
  x[1] = memoire->getWord(adr + 0x10) & 0x7FF;
  y[1] = memoire->getWord(adr + 0x12) & 0x7FF;
  x[2] = memoire->getWord(adr + 0x14) & 0x7FF;
  y[2] = memoire->getWord(adr + 0x16) & 0x7FF;
  x[3] = memoire->getWord(adr + 0x18) & 0x7FF;
  y[3] = memoire->getWord(adr + 0x1A) & 0x7FF;
  unsigned short CMDDGRDA = memoire->getWord(adr + 0x1C);
  filledPolygonColor(surface, x, y, 4, 0xFF0000FF);
  SDL_Flip(surface); // FIXME should not be here
#endif
}

void Vdp1::polylineDraw(void) {
#if DEBUG
  cerr << "vdp1\t: polyline draw" << endl;
#endif
}

void Vdp1::lineDraw(void) {
#if DEBUG
  cerr << "vdp1\t: line draw" << endl;
#endif
}

void Vdp1::userClipping(void) {
#if DEBUG
  cerr << "vdp1\t: user clipping" << endl;
#endif
}

void Vdp1::systemClipping(void) {
#if DEBUG
  cerr << "vdp1\t: system clipping" << endl;
#endif
}

void Vdp1::localCoordinate(void) {
#if DEBUG
  cerr << "vdp1\t: local coordinate" << endl;
#endif
}

void Vdp1::drawEnd(void) {
#if DEBUG
  cerr << "vdp1\t: draw end" << endl;
#endif
  //Scu::sendDrawEnd();
}
